@@ -20,8 +20,12 @@ from nose.tools import assert_equal
 import botocore.session
 from botocore.config import Config
 from botocore.compat import datetime, urlsplit, parse_qs, get_md5
-from botocore.exceptions import ParamValidationError, ClientError
-from botocore.exceptions import InvalidS3UsEast1RegionalEndpointConfigError
+from botocore.exceptions import (
+    ParamValidationError, ClientError,
+    UnsupportedS3ConfigurationError,
+    UnsupportedS3AccesspointConfigurationError,
+    InvalidS3UsEast1RegionalEndpointConfigError,
+)
 from botocore.parsers import ResponseParserError
 from botocore import UNSIGNED
 
@@ -45,10 +49,41 @@ class BaseS3OperationTest(BaseSessionTest):
 
 
 class BaseS3ClientConfigurationTest(BaseSessionTest):
+    _V4_AUTH_REGEX = re.compile(
+        r'AWS4-HMAC-SHA256 '
+        r'Credential=\w+/\d+/'
+        r'(?P<signing_region>[a-z0-9-]+)/'
+        r'(?P<signing_name>[a-z0-9-]+)/'
+    )
+
     def setUp(self):
         super(BaseS3ClientConfigurationTest, self).setUp()
         self.region = 'us-west-2'
 
+    def assert_signing_region(self, request, expected_region):
+        auth_header = request.headers['Authorization'].decode('utf-8')
+        actual_region = None
+        match = self._V4_AUTH_REGEX.match(auth_header)
+        if match:
+            actual_region = match.group('signing_region')
+        self.assertEqual(expected_region, actual_region)
+
+    def assert_signing_name(self, request, expected_name):
+        auth_header = request.headers['Authorization'].decode('utf-8')
+        actual_name = None
+        match = self._V4_AUTH_REGEX.match(auth_header)
+        if match:
+            actual_name = match.group('signing_name')
+        self.assertEqual(expected_name, actual_name)
+
+    def assert_signing_region_in_url(self, url, expected_region):
+        qs_components = parse_qs(urlsplit(url).query)
+        self.assertIn(expected_region, qs_components['X-Amz-Credential'][0])
+
+    def assert_endpoint(self, request, expected_endpoint):
+        actual_endpoint = urlsplit(request.url).netloc
+        self.assertEqual(actual_endpoint, expected_endpoint)
+
     def create_s3_client(self, **kwargs):
         client_kwargs = {
             'region_name': self.region
@@ -468,13 +503,6 @@ class TestS3Copy(BaseS3OperationTest):
 
 
 class TestAccesspointArn(BaseS3ClientConfigurationTest):
-    _V4_AUTH_REGEX = re.compile(
-        r'AWS4-HMAC-SHA256 '
-        r'Credential=\w+/\d+/'
-        r'(?P<signing_region>[a-z0-9-]+)/'
-        r'(?P<signing_name>[a-z0-9-]+)/'
-    )
-
     def setUp(self):
         super(TestAccesspointArn, self).setUp()
         self.client, self.http_stubber = self.create_stubbed_s3_client()
@@ -485,22 +513,6 @@ class TestAccesspointArn(BaseS3ClientConfigurationTest):
         http_stubber.start()
         return client, http_stubber
 
-    def assert_signing_region(self, request, expected_region):
-        auth_header = request.headers['Authorization'].decode('utf-8')
-        actual_region = self._V4_AUTH_REGEX.match(
-            auth_header).group('signing_region')
-        self.assertEqual(expected_region, actual_region)
-
-    def assert_signing_name(self, request, expected_name):
-        auth_header = request.headers['Authorization'].decode('utf-8')
-        actual_name = self._V4_AUTH_REGEX.match(
-            auth_header).group('signing_name')
-        self.assertEqual(expected_name, actual_name)
-
-    def assert_signing_region_in_url(self, url, expected_region):
-        qs_components = parse_qs(urlsplit(url).query)
-        self.assertIn(expected_region, qs_components['X-Amz-Credential'][0])
-
     def assert_expected_copy_source_header(self,
                                            http_stubber, expected_copy_source):
         request = self.http_stubber.requests[0]
@@ -513,10 +525,6 @@ class TestAccesspointArn(BaseS3ClientConfigurationTest):
             body=b'<CopyObjectResult></CopyObjectResult>'
         )
 
-    def assert_endpoint(self, request, expected_endpoint):
-        actual_endpoint = urlsplit(request.url).netloc
-        self.assertEqual(actual_endpoint, expected_endpoint)
-
     def test_missing_region_in_arn(self):
         accesspoint_arn = (
             'arn:aws:s3::123456789012:accesspoint:myendpoint'
@@ -783,9 +791,7 @@ class TestAccesspointArn(BaseS3ClientConfigurationTest):
         )
         self.client, _ = self.create_stubbed_s3_client(
             config=Config(s3={'use_accelerate_endpoint': True}))
-        with self.assertRaises(
-                botocore.exceptions.
-                UnsupportedS3AccesspointConfigurationError):
+        with self.assertRaises(UnsupportedS3AccesspointConfigurationError):
             self.client.list_objects(Bucket=outpost_arn)
 
     def test_outpost_arn_with_s3_dualstack(self):
@@ -795,9 +801,7 @@ class TestAccesspointArn(BaseS3ClientConfigurationTest):
         )
         self.client, _ = self.create_stubbed_s3_client(
             config=Config(s3={'use_dualstack_endpoint': True}))
-        with self.assertRaises(
-                botocore.exceptions.
-                UnsupportedS3AccesspointConfigurationError):
+        with self.assertRaises(UnsupportedS3AccesspointConfigurationError):
             self.client.list_objects(Bucket=outpost_arn)
 
     def test_incorrect_outpost_format(self):
@@ -838,6 +842,54 @@ class TestAccesspointArn(BaseS3ClientConfigurationTest):
         with self.assertRaises(botocore.exceptions.ParamValidationError):
             self.client.list_objects(Bucket=outpost_arn)
 
+    def test_s3_object_lambda_arn_with_s3_dualstack(self):
+        s3_object_lambda_arn = (
+            'arn:aws:s3-object-lambda:us-west-2:123456789012:'
+            'accesspoint/myBanner'
+        )
+        self.client, _ = self.create_stubbed_s3_client(
+            config=Config(s3={'use_dualstack_endpoint': True}))
+        with self.assertRaises(UnsupportedS3AccesspointConfigurationError):
+            self.client.list_objects(Bucket=s3_object_lambda_arn)
+
+    def test_s3_object_lambda_arn_with_us_east_1(self):
+        # test that us-east-1 region is not resolved
+        # into s3 global endpoint
+        s3_object_lambda_arn = (
+            'arn:aws:s3-object-lambda:us-east-1:123456789012:'
+            'accesspoint/myBanner'
+        )
+        self.client, self.http_stubber = self.create_stubbed_s3_client(
+            region_name='us-east-1',
+            config=Config(s3={'use_arn_region': False})
+        )
+        self.http_stubber.add_response()
+        self.client.list_objects(Bucket=s3_object_lambda_arn)
+        request = self.http_stubber.requests[0]
+        self.assert_signing_name(request, 's3-object-lambda')
+        self.assert_signing_region(request, 'us-east-1')
+        expected_endpoint = (
+            'myBanner-123456789012.s3-object-lambda.us-east-1.amazonaws.com'
+        )
+        self.assert_endpoint(request, expected_endpoint)
+
+    def test_basic_s3_object_lambda_arn(self):
+        s3_object_lambda_arn = (
+            'arn:aws:s3-object-lambda:us-west-2:123456789012:'
+            'accesspoint/myBanner'
+        )
+        self.client, self.http_stubber = self.create_stubbed_s3_client(
+            region_name='us-east-1')
+        self.http_stubber.add_response()
+        self.client.list_objects(Bucket=s3_object_lambda_arn)
+        request = self.http_stubber.requests[0]
+        self.assert_signing_name(request, 's3-object-lambda')
+        self.assert_signing_region(request, 'us-west-2')
+        expected_endpoint = (
+            'myBanner-123456789012.s3-object-lambda.us-west-2.amazonaws.com'
+        )
+        self.assert_endpoint(request, expected_endpoint)
+
 
 class TestOnlyAsciiCharsAllowed(BaseS3OperationTest):
     def test_validates_non_ascii_chars_trigger_validation_error(self):
@@ -934,6 +986,59 @@ class TestS3PutObject(BaseS3OperationTest):
             self.assertEqual(len(http_stubber.requests), 2)
 
 
+class TestWriteGetObjectResponse(BaseS3ClientConfigurationTest):
+    def create_stubbed_s3_client(self, **kwargs):
+        client = self.create_s3_client(**kwargs)
+        http_stubber = ClientHTTPStubber(client)
+        http_stubber.start()
+        return client, http_stubber
+
+    def test_endpoint_redirection(self):
+        regions = ['us-west-2', 'us-east-1']
+        for region in regions:
+            self.client, self.http_stubber = self.create_stubbed_s3_client(
+                region_name=region)
+            self.http_stubber.add_response()
+            self.client.write_get_object_response(
+                RequestRoute='endpoint-io.a1c1d5c7',
+                RequestToken='SecretToken',
+            )
+            request = self.http_stubber.requests[0]
+            self.assert_signing_name(request, 's3-object-lambda')
+            self.assert_signing_region(request, region)
+            expected_endpoint = (
+                'endpoint-io.a1c1d5c7.s3-object-lambda.'
+                '%s.amazonaws.com' % region
+            )
+            self.assert_endpoint(request, expected_endpoint)
+
+    def test_endpoint_redirection_fails_with_custom_endpoint(self):
+        self.client, self.http_stubber = self.create_stubbed_s3_client(
+            region_name='us-west-2', endpoint_url="https://example.com")
+        self.http_stubber.add_response()
+        self.client.write_get_object_response(
+            RequestRoute='endpoint-io.a1c1d5c7',
+            RequestToken='SecretToken',
+        )
+        request = self.http_stubber.requests[0]
+        self.assert_signing_name(request, 's3-object-lambda')
+        self.assert_signing_region(request, 'us-west-2')
+        self.assert_endpoint(request, 'endpoint-io.a1c1d5c7.example.com')
+
+    def test_endpoint_redirection_fails_with_accelerate_endpoint(self):
+        config = Config(s3={'use_accelerate_endpoint': True})
+        self.client, self.http_stubber = self.create_stubbed_s3_client(
+            region_name='us-west-2',
+            config=config,
+        )
+        self.http_stubber.add_response()
+        with self.assertRaises(UnsupportedS3ConfigurationError):
+            self.client.write_get_object_response(
+                RequestRoute='endpoint-io.a1c1d5c7',
+                RequestToken='SecretToken',
+            )
+
+
 class TestS3SigV4(BaseS3OperationTest):
     def setUp(self):
         super(TestS3SigV4, self).setUp()
@@ -1001,7 +1106,6 @@ class TestS3SigV4(BaseS3OperationTest):
         self.assertNotIn('content-md5', sent_headers)
 
 
-
 class TestCanSendIntegerHeaders(BaseSessionTest):
 
     def test_int_values_with_sigv4(self):
@@ -2249,8 +2353,13 @@ def _verify_presigned_url_addressing(region, bucket, key, s3_config,
     assert_equal(actual, expected_url)
 
 
-class TestS3DeleteObjects(BaseS3OperationTest):
-    def test_escape_keys_in_xml_payload(self):
+class TestS3XMLPayloadEscape(BaseS3OperationTest):
+    def assert_correct_content_md5(self, request):
+        content_md5_bytes = get_md5(request.body).digest()
+        content_md5 = base64.b64encode(content_md5_bytes)
+        self.assertEqual(content_md5, request.headers['Content-MD5'])
+
+    def test_escape_keys_in_xml_delete_objects(self):
         self.http_stubber.add_response()
         with self.http_stubber:
             response = self.client.delete_objects(
@@ -2262,6 +2371,21 @@ class TestS3DeleteObjects(BaseS3OperationTest):
         request = self.http_stubber.requests[0]
         self.assertNotIn(b'\r\n\r', request.body)
         self.assertIn(b'&#xD;&#xA;&#xD;', request.body)
-        content_md5_bytes = get_md5(request.body).digest()
-        content_md5 = base64.b64encode(content_md5_bytes)
-        self.assertEqual(content_md5, request.headers['Content-MD5'])
+        self.assert_correct_content_md5(request)
+
+    def test_escape_keys_in_xml_put_bucket_lifecycle_configuration(self):
+        self.http_stubber.add_response()
+        with self.http_stubber:
+            response = self.client.put_bucket_lifecycle_configuration(
+                Bucket='mybucket',
+                LifecycleConfiguration={
+                    'Rules': [{
+                        'Prefix': 'my\r\n\rprefix',
+                        'Status': 'ENABLED',
+                    }]
+                }
+            )
+        request = self.http_stubber.requests[0]
+        self.assertNotIn(b'my\r\n\rprefix', request.body)
+        self.assertIn(b'my&#xD;&#xA;&#xD;prefix', request.body)
+        self.assert_correct_content_md5(request)
